#
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.

import datetime
import logging
import os
import shutil
import threading
from PySide2 import QtCore

import paths


class FilesModel(QtCore.QObject):

    ROOT_FILES_DIR = os.path.expanduser("/media")

    INVALID_PATH_CHARS = ("..", "/", "\\")

    STAT_RESULT_PREFIX = "st_"

    on_fileError = QtCore.Signal()
    on_fileList = QtCore.Signal()
    on_diskSpaceLeft = QtCore.Signal()

    def __init__(self, core_model):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self._coreModel = core_model
        self._app = core_model._app
        self._printThread = threading.Thread()
        self._initFilesDir()

    def _initFilesDir(self):
        self._files_dir = self.ROOT_FILES_DIR
        if not os.path.isdir(self._files_dir):
            os.mkdir(self._files_dir)
        os.chdir(self.ROOT_FILES_DIR)

    def getAbsPath(self, path):
        return os.path.abspath(os.path.join(self._files_dir, path))

    def checkNameForHacks(self, name):
        for forbidden in self.INVALID_PATH_CHARS:
            if forbidden in name:
                self.on_fileError.emit("Invalid characters in path:" + forbidden)
                return False
        return True
    
    def _printFile(self, path):
        try:
            with open(self.getAbsPath(path), "rb") as f:
                gcodes = f.read()
        except Exception:
            message = "Exception while opening file to print"
            self.logger.exception(message)
            self.on_fileError.emit(message)
        else:
            self._app.qt_interface.printer_model.loadGcodes(gcodes)

    @QtCore.Slot(str)
    def makeDir(self, dirname):
        if self.checkNameForHacks(dirname):
            try:
                os.mkdir(self.getAbsPath(dirname))
            except Exception as e:
                self.logger.exception("Exception while making directory")
                self.on_fileError.emit(str(e))

    @QtCore.Slot(str)
    def printFile(self, path):
        if self._printThread and self._printThread.is_alive():
            self.logger.error("Error: Another print thread is already exists and is alive")
        else:
            self._printThread = threading.Thread(target=self._printFile, args=[path])
            self._printThread.start()

    @QtCore.Property('QVariantList', notify=on_fileList)
    def changeFilesDirectory(self, path):
        self._files_dir = os.path.join(self._files_dir, path)
        return self.getFilesList()

    @QtCore.Property('QVariantList', notify=on_fileList)
    def filesList(self):
        file_info_list = []
        if self._files_dir != self.ROOT_FILES_DIR:
            file_info_list.append({"filename": "..", "size": None, "atime": "", "mtime": "", "isdir": True})
        try:
            for filename in os.listdir(self._files_dir):
                file_info = {"filename": filename}
                stats = os.stat(filename)
                #file_info["size"] = humanize_file_size(stats.st_size)
                file_info["size"] = stats.st_size
                for attr_name in ("atime", "mtime"):
                    seconds = getattr(stats, self.STAT_RESULT_PREFIX + attr_name)
                    date = datetime.datetime.fromtimestamp(seconds).strftime("%A, %B %d, %Y %I:%M:%S")
                    file_info[attr_name] = date
                file_info["isdir"] = os.path.isdir(filename)
                file_info_list.append(file_info)
        except OSError:
            self.logger.exception("OS error on getting file list:")
        return file_info_list

    @QtCore.Property(str, notify=on_diskSpaceLeft)
    def diskSpaceLeft(self):
        _, _, free_space = shutil.disk_usage("/")
        return paths.humanize_disk_size(free_space)
